from . import errors
from .exception_handler import ExceptionHandler
from threading import Lock

_logger_lock = Lock()


class Thorium(object):

    # The logging format used for the debug logger.  This is only used when
    # the application is in debug mode, otherwise the attached logging
    # handler does the formatting.
    debug_log_format = (
        '-' * 80 + '\n' +
        '%(levelname)s in %(module)s [%(pathname)s:%(lineno)d]:\n' +
        '%(message)s\n' +
        '-' * 80
    )

    exception_cls = ExceptionHandler

    def __init__(self, settings, route_manager, debug=False):
        self._logger = None
        self.logger_name = 'thorium'
        self._settings = settings or {}
        self._route_manager = route_manager
        self.debug_on = debug
        self.exception_handler = self.exception_cls(self.logger)
        self._bind_routes()

    @property
    def logger(self):
        """A :class:`logging.Logger` object for this application.
        """
        if self._logger and self._logger.name == self.logger_name:
            return self._logger
        with _logger_lock:
            if self._logger and self._logger.name == self.logger_name:
                return self._logger
            from .logging import create_logger
            self._logger = rv = create_logger(self.logger_name, self.debug_on, self.debug_log_format)
            return rv


# -*- coding: utf-8 -*-

from . import errors, validators, NotSet


class ResourceField(object):
    validator_type = None
    order_counter = 0

    def __init__(self, default=NotSet, notnull=False, readonly=False,
                 writeonly=False, options=None, cast=None, required=False,
                 *args, **kwargs):
        self.flags = {
            'notnull': notnull,
            'readonly': readonly,
            'default': default,
            'writeonly': writeonly,
            'options': options,
            'cast': cast,
            'required': required,
        }

        self.name = 'noname'

        self.order_value = ResourceField.order_counter
        ResourceField.order_counter += 1

        # a hook to allow subclasses to add their own unique parameters
        self.set_unique_attributes(**kwargs)

        # create validator
        self._validator = self._get_validator()
        self.flags['default'] = self._validator.validate(default, cast=False)
        #self._value = self.flags['default']

    def __str__(self):
        return '{0}:{1}'.format(self.__class__.__name__, self.name)

    def validate(self, value, cast=False):
        cast = self.flags['cast'] if self.flags['cast'] is not None else cast
        return self._validator.validate(value, cast)

    @property
    def is_readonly(self):
        return self.flags['readonly']

    @property
    def is_required(self):
        return self.flags['required']

    @property
    def default(self):
        return self.flags['default']

    def set_unique_attributes(self, **kwargs):
        for key, value in kwargs.items():
            self.flags[key] = value

    def _get_validator(self):
        if self.validator_type:
            vt_cls = self.validator_type
            return vt_cls(self)
        else:
            raise NotImplementedError('Base class ResourceField has no validator')


class CharField(ResourceField):
    validator_type = validators.CharValidator

    def set_unique_attributes(self, max_length=None):
        self.flags['max_length'] = max_length


class IntField(ResourceField):
    validator_type = validators.IntValidator


class DecimalField(ResourceField):
    validator_type = validators.DecimalValidator


class DateField(ResourceField):
    validator_type = validators.DateValidator

    # default cast to True
    def __init__(self, *args, cast=True, **kwargs):
        super().__init__(*args, cast=cast, **kwargs)


class DateTimeField(ResourceField):
    validator_type = validators.DateTimeValidator

    # default cast to True
    def __init__(self, *args, cast=True, **kwargs):
        super().__init__(*args, cast=cast, **kwargs)


class UUIDField(ResourceField):
    validator_type = validators.UUIDValidator

    # default cast to True
    def __init__(self, *args, cast=True, **kwargs):
        super().__init__(*args, cast=cast, **kwargs)


class BoolField(ResourceField):
    validator_type = validators.BoolValidator


class ListField(ResourceField):
    validator_type = validators.ListValidator

    def set_unique_attributes(self, item_type=NotSet):
        if item_type:
            if item_type == NotSet or not isinstance(item_type, ResourceField):
                raise errors.ValidationError('ListField must have an item_type set to a valid ResourceField')
            self.flags['item_type'] = item_type
        else:
            self.flags['item_type'] = None


class DictField(ResourceField):
    validator_type = validators.DictValidator


class SetField(ResourceField):
    validator_type = validators.SetValidator

class JSONField(ResourceField):
    validator_type = validators.JSONValidator

    def set_unique_attributes(self, schema=None):
        if schema:
            self.flags['json_schema'] = schema



"""
    thorium.routing
    ~~~~~~~~~~~~~~~

    This module is a helper for building and holding :class:`.Route` objects before
    they are pulled in by the :class:`.Thorium` object when it initializes.

"""

from . import dispatcher
from .resources import ResourceMetaClass


class Route(object):
    """ An object to hold the relationship between a url path and the
    dispatcher that handles requests to that path.

    Note: To add a route after creation use the :meth:`add_route` method.

    :param name: The name to associate with this route
    :param path: The relative to base url path to access the resource, sometimes referred to as endpoint
    :param dispatcher: A :class:`.Dispatcher` object to handle the request
    """

    def __init__(self, name, path, dispatcher):
        self.name = name
        self.path = path
        self.dispatcher = dispatcher


class RouteManager(object):
    """ Manages the list of :class:`.Route`'s needed by :class:`.Thorium`.
    Register your resources using either :func:`register_resource` or build :class:`.Route`'s
    manually and add them with :func:`add_route`.
    """

    def __init__(self):
        self._routes = []

    def add_route(self, route):
        """
        Adds a :class:`.Route` to internal tracking

        :param route: A :class:`.Route` class to track
        """
        self._routes.append(route)

    def get_all_routes(self):
        """ Returns all :class:`.Route`'s that have been added """

        return self._routes

    def register_endpoint(self, endpoint_cls):
        if not hasattr(endpoint_cls, 'routes') or not isinstance(endpoint_cls.routes, list):
            raise Exception('Endpoint {0} expects attribute routes to be a list of Route objects.'
                            .format(endpoint_cls.__name__))

        for route in endpoint_cls.routes:
            self.add_route(route)


def collection(path, methods, parameters_cls=None):
    def wrapped(cls):
        route = build_route(dispatcher.CollectionDispatcher, cls, parameters_cls, path, methods)
        if hasattr(cls, 'routes'):
            cls.routes.append(route)
        else:
            cls.routes = [route]
        return cls
    return wrapped


def detail(path, methods, parameters_cls=None):
    def wrapped(cls):
        route = build_route(dispatcher.DetailDispatcher, cls, parameters_cls, path, methods)
        if hasattr(cls, 'routes'):
            cls.routes.append(route)
        else:
            cls.routes = [route]
        return cls
    return wrapped


def build_route(dispatcher_cls, endpoint_cls, parameters_cls, path, methods):
    if not hasattr(endpoint_cls, 'Resource') or not isinstance(endpoint_cls.Resource, ResourceMetaClass):
        raise Exception('Endpoint {0} expects attribute resource to have a valid Resource object.'
                        .format(endpoint_cls.__name__))

    dsp = dispatcher_cls(endpoint_cls=endpoint_cls,
                         resource_cls=endpoint_cls.Resource,
                         parameters_cls=parameters_cls,
                         allowed_methods=set(methods))
    route = Route(name='{0}_{1}'.format(endpoint_cls.__name__, dsp.request_type),
                  path=path,
                  dispatcher=dsp)
    return route


# -*- coding: utf-8 -*-

from . import errors
from .response import DetailResponse, CollectionResponse
from .serializer import JsonSerializer


class DispatcherBase(object):
    """
    The Dispatcher holds a :class:`.ResourceInterface` and :class:`.Endpoint`
    pairing and is associated with a :class:`.Route`. It's responsible for
    handing a request off to the correct components.

    :param resource_cls: A :class:`.Resource` class definition
    :param engine: A :class:`.Endpoint` object to implement the
        :class:`.ResourceInterface`
    """

    def __init__(self, endpoint_cls, resource_cls, parameters_cls,
                 allowed_methods):
        self.Endpoint = endpoint_cls
        self.Resource = resource_cls
        self.Parameters = parameters_cls
        self.allowed_methods = {method.upper() for method in allowed_methods}

    def dispatch(self, request):
        """
        Injects the :class:`.ResourceInterface` into the :class:`.Endpoint`
        then calls a method on the engine to handle the request based on the
        :class:`.ThoriumRequest`.

        :param request: A :class:`.ThoriumRequest` object
        """

        # ensure valid method
        if request.method not in self.allowed_methods:
            msg = ('Method {0} not available on {1} {2} resource.'
                   .format(request.method,
                           self.Resource.__name__,
                           self.request_type))
            raise errors.MethodNotAllowedError(
                message=msg, headers={'Allow': ', '.join(self.allowed_methods)}
            )

        response = self.build_response_obj(request=request)

        engine = self.Endpoint(request=request, response=response)

        dispatch_method = self.get_dispatch_method(engine=engine)

        engine.authenticate(dispatch_method)

        engine.pre_request()

        self.pre_request(engine=engine)

        # Call into the endpoint
        method_response = dispatch_method()

        # Override default response
        if method_response:
            response = method_response

        engine.post_request()

        serializer = self.get_serializer()
        serialized_body = serializer.serialize_response(response)

        return response, serialized_body

    def get_dispatch_method(self, engine):
        """ find the method in the engine that matches the request """
        return getattr(engine, "{0}_{1}".format(engine.request.method.lower(),
                                                self.request_type))

    def get_serializer(self):
        return JsonSerializer()


class CollectionDispatcher(DispatcherBase):
    """
    A subclass of :class:`.DispatcherBase` to handle requests made on
    collections of :class:`.Resource`'s
    """
    request_type = 'collection'

    def pre_request(self, engine):
        engine.pre_request_collection()

    def build_response_obj(self, request):
        method = request.method.lower()
        if method == 'post':
            response = DetailResponse(request)
        else:
            response = CollectionResponse(request=request)
        return response


class DetailDispatcher(DispatcherBase):
    """
    A subclass of :class:`.DispatcherBase` to handle requests made on
    individual :class:`.Resource`'s
    """
    request_type = 'detail'

    def pre_request(self, engine):
        engine.pre_request_detail()

    def build_response_obj(self, request):
        return DetailResponse(request)



# -*- coding: utf-8 -*-
"""
    thorium.ext
    ~~~~~~~~~

    :license: BSD, see LICENSE for more details.
"""


def setup():
    from flask.exthook import ExtensionImporter
    importer = ExtensionImporter(['thorium_%s'], __name__)
    importer.install()


setup()
del setup



from . import errors


def use(*authenticator_classes):
    """ A decorator to attach one or more :class:`Authenticator`'s to the decorated class.

    Usage:
        from thorium import auth

        @auth.use(BasicAuth, CustomAuth)
        class MyEngine(Endpoint):
            ...

        OR

        @auth.use(BasicAuth)
        @auth.use(CustomAuth)
        class MyEngine(Endpoint):
            ...

    :param authenticator_classes: One or more :class:`Authenticator` class definitions.
    """
    def wrapped(cls):
        if not cls._authenticator_classes:
            cls._authenticator_classes = []
        cls._authenticator_classes.extend(authenticator_classes)
        return cls
    return wrapped


def validation(func):
    """ A simple decorator which marks a method within an :class:`Authenticator` as a
    validation method. All such methods will be converted into decorators via
    the :class:`AuthenticatorMetaClass`. """
    func.is_validation_method = True
    return func


class AuthenticatorMetaClass(type):
    """ Metaclass for :class:`Authenticator`, finds all validation methods within
    the Authenticator and converts them to classmethod decorators which can be used
    on an :class:`Endpoint`'s methods.
    """

    def __new__(mcs, cls_name, bases, attrs):
        for name, method in attrs.items():
            if hasattr(method, 'is_validation_method'):
                attrs[name] = classmethod(mcs._generate_decorator(method))

        #dynamic creation used so that subclasses don't share these attributes
        attrs['_no_auth_methods'] = set()
        attrs['_validators'] = {}

        return super().__new__(mcs, cls_name, bases, attrs)

    @staticmethod
    def _generate_decorator(validation_method):
        """ Dynamically creates a decorator that will map the decorated
        method to the validation_method.

        :param validation_method: The method with validation rules
        """
        def decorator(cls, func):
            if func not in cls._validators:
                cls._validators[func] = []
            cls._validators[func].append(validation_method)
            return func
        return decorator


class Authenticator(object, metaclass=AuthenticatorMetaClass):
    """ Inherit from this class to create a custom authenticator.

    :param engine: A :class:`Endpoint` object.
    """
    def __init__(self, engine):
        self._loaded = False
        self.engine = engine
        self.request = engine.request
        self.response = engine.response
        self.method = None

    @classmethod
    def no_auth(cls, func):
        """ A special validation decorator which indicates that :function:`authentication`
        should not be called for the decorated method.
        """
        cls._no_auth_methods.add(func)
        return func

    def check_auth(self, method):
        """ Finds and executes all validation rules for the given method.

        :param method: A reference to a class method to run validation on.
        """
        self.method = method.__func__
        if self.method not in self._no_auth_methods:
            self.try_load()
            if not self._authenticate():
                raise errors.UnauthorizedError()

            if not self._authorize():
                raise errors.ForbiddenError()

        if self.method in self._validators:
            self.try_load()
            for v in self._validators[self.method]:
                if not v(self):
                    raise errors.ForbiddenError()

        return True

    def try_load(self):
        """ Call the :function:`load` subclass hook if it hasn't not already been called. """
        if not self._loaded:
            self._load()
            self._loaded = True

    def _load(self):
        """ This method will be called only once immediately before running the first validation method.
        Overriding this method is the preferred way to initiate authenticator data since __init__
        will be called whether validation is required or not. """
        pass

    def _authenticate(self):
        """ This method will always be called from :function:`check_auth` except in the case
        where the method being validated has the :function:`no_auth` decorator. Override to
        provide the default authentication check. A False return will throw an access denied
        error. """
        raise NotImplementedError('No authenticator found.')

    def _authorize(self):
        """ This method will always be called from :function:`check_auth` except in the case
        where the method being validated has the :function:`no_auth` decorator. Override to
        provide the default authorization check. A False return will throw an access denied
        error. """
        raise NotImplementedError('No authorization found.')